package us.kostenko.gps_service_plugin

import android.app.Notification
import android.app.NotificationChannel
import android.app.NotificationManager
import android.app.Service
import android.content.Context
import android.content.Intent
import android.content.SharedPreferences
import android.os.Binder
import android.os.Build
import android.os.IBinder
import android.util.Log
import androidx.annotation.RequiresApi
import com.google.android.gms.location.LocationResult
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.PluginRegistry
import io.flutter.view.FlutterCallbackInformation
import io.flutter.view.FlutterMain
import io.flutter.view.FlutterNativeView
import io.flutter.view.FlutterRunArguments
import us.kostenko.gps_service_plugin.model.Position
import us.kostenko.gps_service_plugin.utils.customPrefs
import us.kostenko.gps_service_plugin.utils.defaultPrefs
import us.kostenko.gps_service_plugin.utils.get
import us.kostenko.gps_service_plugin.utils.set
import us.kostenko.gps_service_plugin.utils.toDateString
import java.util.ArrayDeque
import java.util.concurrent.atomic.AtomicBoolean

class LocationService : Service(), MethodChannel.MethodCallHandler {

    private val binder = LocationServiceBinder()
    private var notificationManager: NotificationManager? = null
    private lateinit var locationListener: LocationListener

    private val queue = ArrayDeque<Map<String, Any?>>()
    private lateinit var mBackgroundChannel: MethodChannel

    override fun onBind(intent: Intent?): IBinder? = binder

    fun startSession() {
        locationListener.startTrecking()
        defaultPrefs[IS_TRECKING_STARTED] = true
        lifecycleEvent("startTrecking")
    }

    fun stopSession() {
        locationListener.stopTracking()
        defaultPrefs[IS_TRECKING_STARTED] = false
        lifecycleEvent("stopTrecking")
    }

    private fun startGpsService(context: Context) {
        // Synchronize on sServiceStarted to avoid multiple concurrent initializations.
        synchronized(sServiceStarted) {
            // If we don't have an existing background FlutterNativeView, we need
            // to create one and have it initialize our callback dispatcher.
            if (sBackgroundFlutterView == null) {
                // Grab the callback handle for the callback dispatcher from storage.
                val prefs: SharedPreferences = context.customPrefs(GpsServicePlugin.SHARED_PREFERENCES_KEY)
                Log.i(TAG, "Starting LocationService..., prefs: $prefs")
                val callbackHandle = prefs[GpsServicePlugin.CALLBACK_DISPATCHER_HANDLE_KEY] ?: 0L

                Log.i(TAG, "Starting LocationService..., callbackHandle: $callbackHandle")
                // Retrieve the actual callback information needed to invoke it.
                val callbackInfo = FlutterCallbackInformation.lookupCallbackInformation(callbackHandle)
                Log.i(TAG, "Starting LocationService..., callbackHandle: $callbackHandle, callbackInfo: $callbackInfo")

                // Create a FlutterNativeView with no drawing surface (i.e., a background view).
                sBackgroundFlutterView = FlutterNativeView(context, true)

                // Register all plugins for the application with our new
                // FlutterNativeView's plugin registry. Other plugins will not
                // work when running in the background if this isn't done!
                val registry = sBackgroundFlutterView!!.pluginRegistry
                sPluginRegistrantCallback.registerWith(registry)

                val args = FlutterRunArguments()
                args.bundlePath = FlutterMain.findAppBundlePath(context)
                args.entrypoint = callbackInfo.callbackName
                args.libraryPath = callbackInfo.callbackLibraryPath

                // Start running callback dispatcher code in our background view.
                sBackgroundFlutterView!!.runFromBundle(args)
            }
        }
        mBackgroundChannel = MethodChannel(sBackgroundFlutterView, "us.kostenko/gps_plugin_background")
        mBackgroundChannel.setMethodCallHandler(this)
    }

    override fun onCreate() {
        super.onCreate()

        locationListener = LocationListener(this, :: onLocationResult)

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            // Api 26+ requires running in foreground to keep service alive
            // https://stackoverflow.com/a/50691494/4656400
            startForeground(12345678, getNotification())
        }

        startGpsService(this)
    }

    override fun onTrimMemory(level: Int) {
        super.onTrimMemory(level)
        lifecycleEvent("onTrimMemory, level: $level")
    }

    override fun onStartCommand(intent: Intent?, flags: Int, startId: Int): Int {
        super.onStartCommand(intent, flags, startId)

        if (intent == null && isTreckingLocation) { // intent == null indicates that service was killed by system and was recreated
            locationListener.startTrecking()
        }

        lifecycleEvent("onStartCommand, intent: $intent")

        return START_STICKY
    }

    override fun onDestroy() {
        super.onDestroy()
        lifecycleEvent("onDestroy")
    }

    /** MethodChannel.MethodCallHandler */

    override fun onMethodCall(call: MethodCall, result: MethodChannel.Result) {
        when(call.method) {
            "LocationService.initialized" -> {
                synchronized(sServiceStarted) {
                    while (!queue.isEmpty()) {
                        mBackgroundChannel.invokeMethod("", queue.remove())
                    }
                    sServiceStarted.set(true)
                }
                result.success(null)
            }
            else -> result.notImplemented()
        }
    }

    /** Private methods */

    private fun handlePosition(position: Position) {

        val locationMap: Map<String, Any?> = mapOf(
                "time"      to position.time,       // string
                "latitude"  to position.latitude,   // double
                "longitude" to position.longitude,  // double
                "accuracy"  to position.accuracy,   // float
                "altitude"  to position.altitude,   // double
                "speed"     to position.speed       // float
        )

        val gpsUpdateMap: Map<String, Any?> = mapOf(
                "type"              to eventTypePosition,
                "callbackHandle"    to getCallbackHandle(),       // string
                "locationMap"       to locationMap   // double
        )

        handleUpdate(gpsUpdateMap)
    }

    private fun handleUpdate(updateMap: Map<String, Any?>) {

        synchronized(sServiceStarted) {
            if (!sServiceStarted.get()) {
                // Queue up geofencing events while background isolate is starting
                queue.add(updateMap)
            } else {
                // Callback method name is intentionally left blank.
                mBackgroundChannel.invokeMethod("", updateMap)
            }
        }
    }

    private fun lifecycleEvent(event: String) {
        Log.i(TAG, event)

        val lifecycleUpdateMap: Map<String, Any?> = mapOf(
                "type"              to eventTypeLyfecycle,
                "callbackHandle"    to getCallbackHandle(),
                "event"             to event
        )

        handleUpdate(lifecycleUpdateMap)
    }

    private fun getCallbackHandle(): Long {
        val prefs: SharedPreferences = customPrefs(GpsServicePlugin.SHARED_PREFERENCES_KEY)
        val callbackHandle = prefs[GpsServicePlugin.CALLBACK_HANDLE_KEY] ?: 0L
        Log.d(TAG, "LocationService, callbackHandle: $callbackHandle")
        return callbackHandle
    }

    private fun onLocationResult(locationResult: LocationResult?) {
        Log.d(TAG, "locations size: ${locationResult?.locations?.size}")
        locationResult?.locations?.forEach { loc ->
            val position = Position(
                    time = loc.time.toDateString(this),
                    latitude = loc.latitude,
                    longitude = loc.longitude,
                    accuracy = loc.accuracy,
                    altitude = loc.altitude,
                    speed = loc.speed
            )
            handlePosition(position)
        }
    }

    @RequiresApi(Build.VERSION_CODES.O)
    private fun getNotification(): Notification {
        val channelId = "gps_plugin_channel"
        val channel = NotificationChannel(channelId, "Gps Plugin", NotificationManager.IMPORTANCE_LOW)
        val imageId = resources.getIdentifier("ic_launcher", "mipmap", packageName)

        notificationManager = getSystemService(NotificationManager::class.java)
        notificationManager?.createNotificationChannel(channel)
        return Notification.Builder(applicationContext, channelId)
                .setContentTitle("Flutter Gps Test")
                .setContentText("Foreground service is running")
                .setSmallIcon(imageId)
                .build()
    }

    private val isTreckingLocation: Boolean
        get() = defaultPrefs[IS_TRECKING_STARTED] ?: false

    inner class LocationServiceBinder : Binder() {
        val service: LocationService
            get() = this@LocationService
    }

    companion object {

        @JvmStatic private val eventTypePosition = 0

        @JvmStatic private val eventTypeLyfecycle = 1
        
        @JvmStatic val IS_TRECKING_STARTED = "isTreckingStarted"

        @JvmStatic private var sBackgroundFlutterView: FlutterNativeView? = null

        @JvmStatic private val sServiceStarted = AtomicBoolean(false)

        @JvmStatic val TAG: String = LocationService::class.java.simpleName

        @JvmStatic private lateinit var sPluginRegistrantCallback: PluginRegistry.PluginRegistrantCallback

        @JvmStatic fun setPluginRegistrant(callback: PluginRegistry.PluginRegistrantCallback) {
            sPluginRegistrantCallback = callback
        }
    }

}